package drzewa;
import java.awt.*;
import java.awt.event.*;

/**
 *Podstawowa klasa reprezentujaca drzewo AVL
 */
public class DrzewoAVL implements IKrok,ActionListener {
    static final int OdstepX = 10;
    static final int OdstepY = 15;
    Container container; // tu maja lezec wiercholki
    Wierzcholek w; // wierzcholek w korzeniu drzewa
    DrzewoAVL d1,d2,parent; // dzieci drzewa
    int h; // wysokosc drzewa (w sensie drzewowym)
    
    static int GetH(DrzewoAVL d) { if(d == null) return 0; return d.h; }
    
    void chkH() { // przelicza h i aktualizuje. (idzie rekurencyjnie w gore!)
        h = C.max(GetH(d1),GetH(d2)) + 1;
        if(parent == null) return;
        parent.chkH();
    }
    
    /**
     *zwraca wartosc w korzeniu
     */
    public int getVal() { return w.getVal(); }
    
    /**
     *ustawia wartosc w korzeniu
     */
    public void setVal(int val) {
        w.setVal(val);
        container.repaint();
    }
    
    /**
     *zwraca korzen
     */
    public Wierzcholek getW() { return w; }
    
    /**
     *zmienia korzen
     */
    public void setW(Wierzcholek aw) { // aw musi juz lezec na tym containerze co trzeba(todo: zapewnic to)
        if(w != null) {
            aw.pos = w.pos;
            C.NoActionListeners(w);
        }
        C.NoActionListeners(aw);
        w = aw;
        w.addActionListener(this);
    }
    
    /**
     *zwraca pierwsze dziecko
     */
    public DrzewoAVL getD1() { return d1; }
    
    /**
     *zwraca drugie dziecko
     */
    public DrzewoAVL getD2() { return d2; }
    
    /**
     *zmienia pierwsze dziecko
     */
    public void setD1(DrzewoAVL d) {
        d1 = d;
        if(d1 != null) d1.parent = this;
        setPosD1();
        chkH();
    }
    
    /**
     *zmienia drugie dziecko
     */
    public void setD2(DrzewoAVL d) {
        d2 = d;
        if(d2 != null) d2.parent = this;
        setPosD2();
        chkH();
    }
    
    
    public DrzewoAVL(Container acontainer,int val,DrzewoAVL ad1,DrzewoAVL ad2) { init(acontainer,val,ad1,ad2); }
    public DrzewoAVL(Container acontainer,int val) { init(acontainer,val,null,null); }
    public DrzewoAVL(Container acontainer,Wierzcholek aw,DrzewoAVL ad1,DrzewoAVL ad2) { init(acontainer,aw,ad1,ad2); }
    // aw musi juz lezec na tym containerze co trzeba(todo: zapewnic to)
    public DrzewoAVL(Container acontainer,Wierzcholek aw) { init(acontainer,aw,null,null); }
    // aw musi juz lezec na tym containerze co trzeba(todo: zapewnic to)
    void init(Container acontainer,Wierzcholek aw,DrzewoAVL ad1,DrzewoAVL ad2) {
        // aw musi juz lezec na tym containerze co trzeba(todo: zapewnic to)
        container = acontainer;
        setW(aw);
        setD1(ad1);
        setD2(ad2);
    }
    void init(Container acontainer,int val,DrzewoAVL ad1,DrzewoAVL ad2) {
        Wierzcholek aw = new Wierzcholek(val);
        acontainer.add(aw);
        aw.setSize(aw.getMinimumSize());
        init(acontainer,aw,ad1,ad2);
        container.repaint();
    }
    
    static int GetWidth(DrzewoAVL d) { if(d == null) return 0; return d.getWidth(); }
    
    /**
     *zwraca szerokosc drzewa (w pixelach)
     */
    public int getWidth() {
        int wl = C.max(GetWidth(d1)+OdstepX/2,w.getWidth()/2);
        int wr = C.max(GetWidth(d2)+OdstepX/2,w.getWidth()/2);
        return wl + wr;
    }
    
    static int GetHeight(DrzewoAVL d) { if(d == null) return 0; return d.getHeight(); }
    
    /**
     *zwraca wysokosc drzewa (w pixelach)
     */
    public int getHeight() {
        return w.getHeight() + OdstepY + C.max(GetHeight(d1),GetHeight(d2));
    }
    
    /**
     *przesowa drzewo
     *@param p pozycja lewego gornego rogu obszaru zajetego przez drzewo
     */
    public void setLocation(Point p) {
        setPos(new Point(p.x + C.max(GetWidth(d1) + OdstepX/2,w.getWidth()/2),p.y + w.getHeight()/2));
    }
    
    /**
     *zwraca pozycje drzewa (jako pozycja korzenia)
     */
    public Point getPos() { return w.pos; }
    
    /**
     *przesowa drzewo dla zadanej pozycji korzenia
     */
    public void setPos(Point p) {
        w.pos = p;
        setPosD1();
        setPosD2();
    }
    
    /**
     *poprawia pozycje lewego dziecka
     */
    public void setPosD1() {
        if(d1 == null) return;
        int x = w.pos.x - OdstepX/2 - d1.getWidth();
        int y = w.pos.y + w.getHeight()/2 + OdstepY;
        d1.setLocation(new Point(x,y));
    }
    
    /**
     *poprawia pozycje prawego dziecka
     */
    public void setPosD2() { // poprawia pozycje d2
        if(d2 == null) return;
        int x = w.pos.x + OdstepX/2;
        int y = w.pos.y + w.getHeight()/2 + OdstepY;
        d2.setLocation(new Point(x,y));
    }
    
    /**
     *rysuje krawedzie na podanej grafice
     *( to powinno byc wywolywane w container.paint(Graphics) )
     */
    public void paint(Graphics g) {
        if(d1 != null) {
            C.DrawLine(g,w.getAktPos(),d1.w.getAktPos());
            d1.paint(g);
        }
        if(d2 != null) {
            C.DrawLine(g,w.getAktPos(),d2.w.getAktPos());
            d2.paint(g);
        }
    }
    
    /**
     *robi krok kazdym wierzcholkiem drzewa
     */
    public void krok() {
        w.krok();
        if(d1 != null) d1.krok();
        if(d2 != null) d2.krok();
        container.repaint();
    }
    
    /**
     *wykonuje pojedyncza rotacje w gore
     *uwaga: to NISZCZY obiekt this (wiec trzeba o nim zapimniec!)
     *ale parenta nie niszczy (tylko modyfikuje troche)
     */
    public void rotacja() {
        if(parent == null) return;
        DrzewoAVL pd1 = parent.d1;
        DrzewoAVL pd2 = parent.d2;
        Wierzcholek wp = parent.w;
        parent.setW(w);
        if(pd1 == this) {
            parent.setD1(d1);
            parent.setD2(new DrzewoAVL(container,wp,d2,pd2));
        } else {
            parent.setD1(new DrzewoAVL(container,wp,pd1,d1));
            parent.setD2(d2);
        }
        
                /*
                // jeszcze ewentualne poprawienie rozstawienia (dziadka)
                if(parent.parent != null) {
                        if(parent == parent.parent.d1) parent.parent.setPosD1();
                        else parent.parent.setPosD2();
                }
                 */
        
    }
    
    /**
     *klikniecie na wierzcholek powoduje rotacje
     */
    public void actionPerformed(ActionEvent e) {
        rotacja();
    }
    
    /**
     *wstawia wierzcholek do drzewa (jako lisc na odp. miejscu - z zachowaniem porzadku)
     *zwraca to miejsce w ktore wstawil (albo null jak juz bylo)
     */
    public DrzewoAVL wstaw(Wierzcholek aw) {
        if(aw.getVal() == w.getVal()) return null;
        DrzewoAVL d;
        if(aw.getVal() <  w.getVal()) {
            if(d1 != null) return d1.wstaw(aw);
            d = new DrzewoAVL(container,aw);
            setD1(d);
        } else {
            if(d2 != null) return d2.wstaw(aw);
            d = new DrzewoAVL(container,aw);
            setD2(d);
        }
        return d;
    }
    
    /**
     *jak wyzej ale samo se tworzy wierzcholek z podana wartoscia
     */
    public DrzewoAVL wstaw(int val) {
        if(val == w.getVal()) return null;
        if(val <  w.getVal()) if(d1 != null) return d1.wstaw(val);
        if(val >  w.getVal()) if(d2 != null) return d2.wstaw(val);
        Wierzcholek aw = new Wierzcholek(val);
        container.add(aw);
        aw.setSize(aw.getMinimumSize());
        DrzewoAVL d = wstaw(aw);
        container.repaint();
        return d;
    }
    
    /**
     *jak to wywolamy dla drzewa zwroconego przez 'wstaw' to naprawi rownowage AVL
     */
    public void chkAVL() {
        if(parent == null) return;
        if(isOK()) {
            parent.chkAVL();
            return;
        }
        // teraz dwa symetryczne przypadki:
        if(this == parent.d1) { // pierwszy symatryczny przypadek:
            // i dwa podprzypadki:
            if(GetH(d1) > GetH(d2)) { // pierwszy podprzypadek (prostszy):
                rotacja();
            } else { // drugi podprzypadek (trudniejszy):
                d2.rotacja();
                rotacja();
            }
        } else { // drugi symetryczny przypadek:
            // i znowu dwa podprzypadki:
            if(GetH(d1) < GetH(d2)) { // pierwszy podprzypadek (prostszy):
                rotacja();
            } else { // drugi podprzypadek (trudniejszy):
                d1.rotacja();
                rotacja();
            }
        }
        
    }
    
    /**
     *sprawdza czy this nie jest za dlugi w stosunku do swojego brata
     */
    boolean isOK() {
        if(parent == null) return true;
        int bh; // wysokosc brata
        if(this == parent.d1) bh = GetH(parent.d2);
        else bh = GetH(parent.d1);
        return h - bh < 2;
    }
}

